@@ -1,9 +1,11 @@
 from django import forms
 from django.contrib.contenttypes.models import ContentType
+from django.contrib.admin.models import LogEntry, ADDITION, CHANGE
 from django.contrib.admin.views.decorators import staff_member_required
+from django.conf import settings
 from django.core.exceptions import SuspiciousOperation
 from django.core.urlresolvers import reverse
-from django.db.models import Q
+from django.db.models import Q, ForeignKey
 from django.db import transaction
 from django.db import IntegrityError
 from django.core.exceptions import ObjectDoesNotExist
@@ -13,30 +15,47 @@ from django.shortcuts import render_to_response, get_object_or_404
 from django.template import RequestContext
 import sys
 
-from simple_import.models import ImportLog, ImportSetting, ColumnMatch, ImportedObject
-from simple_import.forms import ImportForm, MatchForm
+from simple_import.models import ImportLog, ImportSetting, ColumnMatch, ImportedObject, RelationalMatch
+from simple_import.forms import ImportForm, MatchForm, MatchRelationForm
 
-def validate_match_columns(import_log, field_names, model_class, header_row):
+def validate_match_columns(import_log, model_class, header_row):
     """ Perform some basic pre import validation to make sure it's
     even possible the import can work
     Returns list of errors
     """
     errors = []
     column_matches = import_log.import_setting.columnmatch_set.all()
+    field_names = model_class._meta.get_all_field_names()
     for field_name in field_names:
         field_object, model, direct, m2m = model_class._meta.get_field_by_name(field_name)
-        if (direct and
-            model and
-            not field_object.blank):
+        if direct and model and not field_object.blank:
             field_matches = column_matches.filter(field_name=field_name)
+            match_in_header = False
             if field_matches:
-                if field_matches[0].column_name not in header_row:
+                for field_match in field_matches:
+                    if field_match.column_name in header_row:
+                        match_in_header = True
+                if not match_in_header:
                     errors += ["{0} is required but is not in your spreadsheet. ".format(field_object.verbose_name)]
             else:
                 errors += ["{0} is required but has no match.".format(field_object.verbose_name)]
     
     return errors
 
+
+def get_custom_fields_from_model(model_class):
+    """ django-custom-fields support
+    """
+    if 'custom_field' in settings.INSTALLED_APPS:
+        from custom_field.models import CustomField
+        try:
+            content_type = ContentType.objects.get(model=model_class._meta.module_name,app_label=model_class._meta.app_label)
+        except ContentType.DoesNotExist:
+            content_type = None
+        custom_fields = CustomField.objects.filter(content_type=content_type)
+        return custom_fields
+    
+    
 @staff_member_required
 def match_columns(request, import_log_id):
     """ View to match import spreadsheet columns with database fields
@@ -74,7 +93,6 @@ def match_columns(request, import_log_id):
                     errors += ['Please select an update key. This key is used to linked records for updating.']
             errors += validate_match_columns(
                 import_log,
-                field_names,
                 model_class,
                 header_row)
             all_field_names = []
@@ -84,47 +102,46 @@ def match_columns(request, import_log_id):
                         errors += ["{0} is duplicated.".format(clean_data['field_name'])]
                     all_field_names += [clean_data['field_name']]
             if not errors:
-                get = ''
-                if 'commit' in request.POST:
-                    get = "?commit=True"
                 return HttpResponseRedirect(reverse(
-                    do_import,
-                    kwargs={'import_log_id': import_log.id}) + get)
+                    match_relations,
+                    kwargs={'import_log_id': import_log.id}))
     else:
-        match_ids = []
-        for cell in header_row:
-            try:
-                match = ColumnMatch.objects.get(
-                    import_setting = import_log.import_setting,
-                    column_name = cell,
-                )
-            except ColumnMatch.DoesNotExist:
-                match = ColumnMatch(
-                    import_setting = import_log.import_setting,
-                    column_name = cell,
-                )
-                match.guess_field()
-                match.save()
-            match_ids += [match.id]
-        
-        existing_matches = ColumnMatch.objects.filter(id__in=match_ids)
+        existing_matches = import_log.get_matches()
         formset = MatchFormSet(queryset=existing_matches)
         
     field_choices = (('', 'Do Not Use'),)
     for field_name in field_names:
         field_object, model, direct, m2m = model_class._meta.get_field_by_name(field_name)
+        add = True
         
         if direct:
             field_verbose = field_object.verbose_name
         else:
             field_verbose = field_name
         
-        if direct and model and not field_object.blank:
+        if direct and  not field_object.blank:
             field_verbose += " (Required)"
         if direct and field_object.unique:
             field_verbose += " (Unique)"
+        if m2m or isinstance(field_object, ForeignKey):
+            field_verbose += " (Related)"
+        elif not direct:
+            add = False
         
-        field_choices += ((field_name, field_verbose),)
+        if add:
+            field_choices += ((field_name, field_verbose),)
+    
+    # Include django-custom-field support
+    custom_fields = get_custom_fields_from_model(model_class)
+    for custom_field in custom_fields:
+        field_choices += (("simple_import_custom__{0}".format(custom_field),
+                           "{0} (Custom)".format(custom_field)),)
+    # Include defined methods
+    # Model must have a simple_import_methods defined
+    if hasattr(model_class, 'simple_import_methods'):
+        for import_method in model_class.simple_import_methods:
+            field_choices += (("simple_import_method__{0}".format(import_method),
+                               "{0} (Method)".format(import_method)),)
     
     i = 0
     for form in formset:
@@ -138,6 +155,91 @@ def match_columns(request, import_log_id):
         RequestContext(request, {}),)
 
 
+def get_direct_fields_from_model(model_class):
+    direct_fields = []
+    all_fields_names = model_class._meta.get_all_field_names()
+    for field_name in all_fields_names:
+        field = model_class._meta.get_field_by_name(field_name)
+        # Direct, not m2m, not FK
+        if field[2] and not field[3] and field[0].__class__.__name__ != "ForeignKey":
+            direct_fields += [field[0]]
+    return direct_fields
+
+
+@staff_member_required
+def match_relations(request, import_log_id):
+    import_log = get_object_or_404(ImportLog, id=import_log_id)
+    model_class = import_log.import_setting.content_type.model_class()
+    matches = import_log.get_matches()
+    field_names = []
+    choice_set = []
+    for match in matches.exclude(field_name=""):
+        if (not match.field_name.startswith('simple_import_custom__') and
+            not match.field_name.startswith('simple_import_method__')):
+            field, model, direct, m2m = model_class._meta.get_field_by_name(match.field_name)
+            if m2m or isinstance(field, ForeignKey): 
+                RelationalMatch.objects.get_or_create(
+                    import_log=import_log,
+                    field_name=match.field_name)
+                field_names += [match.field_name]
+                choices = (('', '---------'),)
+                for field in get_direct_fields_from_model(field.related.parent_model()):
+                    if field.unique:
+                        choices += ((field.name, field.verbose_name),)
+                choice_set += [choices]
+    
+    existing_matches = RelationalMatch.objects.filter(
+        import_log=import_log,
+        field_name__in=field_names)
+    MatchRelationFormSet = modelformset_factory(
+        RelationalMatch,
+        form=MatchRelationForm, extra=0)
+    if request.POST:
+        formset = MatchRelationFormSet(request.POST)
+        if formset.is_valid():
+            formset.save()
+            get = ''
+            if 'commit' in request.POST:
+                get = "?commit=True"
+            return HttpResponseRedirect(reverse(
+                do_import,
+                kwargs={'import_log_id': import_log.id}) + get)
+    else:
+        formset = MatchRelationFormSet(queryset=existing_matches)
+    
+    for i, form in enumerate(formset.forms):
+        choices = choice_set[i]
+        form.fields['related_field_name'].widget = forms.Select(choices=choices)
+        
+    return render_to_response(
+        'simple_import/match_relations.html',
+        {'formset': formset,
+         'existing_matches': existing_matches},
+        RequestContext(request, {}),)
+
+def set_field_from_cell(import_log, new_object, header_row_field_name, cell):
+    """ Set a field from a import cell. Use referenced fields the field
+    is m2m or a foreign key.
+    """
+    if (not header_row_field_name.startswith('simple_import_custom__') and
+            not header_row_field_name.startswith('simple_import_method__')):
+        field, model, direct, m2m =  new_object._meta.get_field_by_name(header_row_field_name)
+        if m2m:
+            new_object.simple_import_m2ms[header_row_field_name] = cell
+        elif isinstance(field, ForeignKey):
+            related_field_name = RelationalMatch.objects.get(import_log=import_log, field_name=field.name).related_field_name
+            related_model = field.related.parent_model
+            related_object = related_model.objects.get(**{related_field_name:cell})
+            setattr(new_object, header_row_field_name, related_object)
+        else:
+            setattr(new_object, header_row_field_name, cell)
+    elif header_row_field_name.startswith('simple_import_custom__'):
+        new_object.set_custom_value(header_row_field_name[22:], cell)
+    elif header_row_field_name.startswith('simple_import_method__'):
+        if header_row_field_name[22:] in new_object.simple_import_methods:
+            getattr(new_object, header_row_field_name[22:])(cell)
+       
+
 @staff_member_required
 def do_import(request, import_log_id):
     """ Import the data!
@@ -159,6 +261,7 @@ def do_import(request, import_log_id):
     header_row = import_data.pop(0)
     header_row_field_names = []
     header_row_default = []
+    header_row_null_on_empty = []
     error_data = [header_row + ['Error Type', 'Error Details']]
     create_count = 0
     update_count = 0
@@ -170,11 +273,14 @@ def do_import(request, import_log_id):
     
     key_column_name = None
     if import_log.update_key and import_log.import_type in ["U", "O"]:
-        key_column_name = import_log.import_setting.columnmatch_set.get(field_name=import_log.update_key).column_name
+        key_match = import_log.import_setting.columnmatch_set.get(column_name=import_log.update_key)
+        key_column_name = key_match.column_name
+        key_field_name = key_match.field_name
     for i, cell in enumerate(header_row):
         match = import_log.import_setting.columnmatch_set.get(column_name=cell)
         header_row_field_names += [match.field_name]
         header_row_default += [match.default_value]
+        header_row_null_on_empty += [match.null_on_empty]
         if key_column_name == cell:
             key_index = i
     
@@ -185,25 +291,49 @@ def do_import(request, import_log_id):
                 if import_log.import_type == "N":
                     new_object = model_class()
                 elif import_log.import_type == "O":
-                    filters = {import_log.update_key: row[key_index]}
+                    filters = {key_field_name: row[key_index]}
                     new_object = model_class.objects.get(**filters)
                     is_created = False
                 elif import_log.import_type == "U":
-                    filters = {import_log.update_key: row[key_index]}
+                    filters = {key_field_name: row[key_index]}
                     try:
                         new_object = model_class.objects.get(**filters)
                         is_created = False
                     except model_class.DoesNotExist:
                         new_object = model_class()
+                new_object.simple_import_m2ms = {} # Need to deal with these after saving
                 for i, cell in enumerate(row):
-                    if cell:
-                        setattr(new_object, header_row_field_names[i], cell)
+                    if cell or header_row_null_on_empty[i]:
+                        set_field_from_cell(import_log, new_object, header_row_field_names[i], cell)
                     elif header_row_default[i]:
-                        setattr(new_object, header_row_field_names[i], header_row_default[i])
+                        set_field_from_cell(import_log, new_object, header_row_field_names[i], header_row_default[i])
                 new_object.save()
+                
+                for key in new_object.simple_import_m2ms.keys():
+                    value = new_object.simple_import_m2ms[key]
+                    m2m = getattr(new_object, key)
+                    m2m_model = type(m2m.model())
+                    related_field_name = RelationalMatch.objects.get(import_log=import_log, field_name=key).related_field_name
+                    m2m_object = m2m_model.objects.get(**{related_field_name:value})
+                    m2m.add(m2m_object)
+                
                 if is_created:
+                    LogEntry.objects.log_action(
+                        user_id         = request.user.pk, 
+                        content_type_id = ContentType.objects.get_for_model(new_object).pk,
+                        object_id       = new_object.pk,
+                        object_repr     = unicode(new_object), 
+                        action_flag     = ADDITION
+                    )
                     create_count += 1
                 else:
+                    LogEntry.objects.log_action(
+                        user_id         = request.user.pk, 
+                        content_type_id = ContentType.objects.get_for_model(new_object).pk,
+                        object_id       = new_object.pk,
+                        object_repr     = unicode(new_object), 
+                        action_flag     = CHANGE
+                    )
                     update_count += 1
                 ImportedObject.objects.create(
                     import_log = import_log,
